package dataStructures.utils;

import java.text.Format;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.math.NumberUtils;

public final class ArrayS
{
	public static final float PERCENTAGE_SHUFFLE_VERTEX_ORDER = 0.1f;

	/**
	 * Simple BucketSort implementation (ascending order) that runs in O(n) time to sort n elements when each of which has a key of value O(n).
	 * 
	 * @param arr
	 * @param initArraySize optional parameter (Default: -1) that might be used to reallocate a larger bucket array during process
	 */
	public static <T extends Key> void bucketSort(final ArrayList<T> arr, final int initArraySize)
	{
		final ArrayList<LinkedList<T>> buckets = new ArrayList<LinkedList<T>>(Math.max(initArraySize, 10));

		for (final T element : arr)
		{
			{// ensure that the bucket for {@code element} can be accessed
				final int minCapacity = element.getComparisonKey() + 1;
				buckets.ensureCapacity(minCapacity);
				while (minCapacity > buckets.size())
				{
					buckets.add(null);
				}
			}
			LinkedList<T> list = buckets.get(element.getComparisonKey());
			if (list == null)// bucket empty?
			{
				list = new LinkedList<T>();
				buckets.set(element.getComparisonKey(), list);
			}
			list.add(element);
		}
		int i = 0;
		for (final LinkedList<T> bucket : buckets)
		{
			if (bucket != null)
			{
				for (final T element : bucket)
				{
					arr.set(i++, element);
				}
			}
		}

	}

	public static float max(final float[] arr)
	{
		return NumberUtils.max(arr);
	}

	public static int max(final Integer[] arr)
	{
		return NumberUtils.max(ArrayUtils.toPrimitive(arr));
	}

	public static double mean(final double[] greedyK)
	{
		double mean = 0;
		for (final double element : greedyK)
		{
			mean += element;
		}
		return mean / greedyK.length;
	}

	public static float mean(final float[] greedyTimeA)
	{
		float mean = 0;
		for (final float element : greedyTimeA)
		{
			mean += element;
		}
		return mean / greedyTimeA.length;
	}

	public static double mean(final int[] greedyK)
	{
		double mean = 0;
		for (final int element : greedyK)
		{
			mean += element;
		}
		return mean / greedyK.length;
	}

	/**
	 * Returns the index of the median of the three indexed integers.
	 */
	private static int med3(final int x[], final int[] p, final int a, final int b, final int c)
	{
		return x[p[a]] < x[p[b]] ? x[p[b]] < x[p[c]] ? b : x[p[a]] < x[p[c]] ? c : a : x[p[b]] > x[p[c]] ? b : x[p[a]] > x[p[c]] ? c : a;
	}

	public static ArrayList<Integer> randomPermutation(final Set<Integer> keySet)
	{
		final ArrayList<Integer> dataRecords = new ArrayList<Integer>(keySet);
		for (int step = 1; step < dataRecords.size(); step++)
		{
			for (int i = 0; (i + step) < dataRecords.size(); i++)
			{
				if (Math.random() <= 0.5f)
				{
					final Integer tmp = dataRecords.get(i);
					dataRecords.set(i, dataRecords.get(i + step));
					dataRecords.set(i + step, tmp);
				}
			}
		}
		return dataRecords;
	}

	/**
	 * 
	 * @param list sorted List
	 */
	public static <T extends Comparable<T>> void shuffleEqualSortKeyElements(final List<T> list)
	{
		final int size = list.size();
		for (int i = 0; i < (size - 1); i++)
		{
			final int begin = i;
			final int end;// index of first element different from the ith element

			{// set end value
				int step = 1;

				while (list.get(begin).compareTo(list.get(begin + step)) == 0)
				{
					step = step << 1;
					if ((step + begin) >= size)
					{
						break;
					}
				}
				int lB = begin + (step >> 1);
				int uB = Math.min(begin + step, size);
				while (lB < (uB - 1))
				{
					final int testKey = lB + ((uB - lB) >> 1);
					if (list.get(begin).compareTo(list.get(testKey)) == 0)
					{
						lB = testKey;
					} else
					{
						uB = testKey;
					}

				}

				end = uB;
			}
			i = end - 1;

			Collections.shuffle(list.subList(begin, end));
		}
	}

	public static <Elements> void shuffleEqualSortKeyElements(final Vector<Elements> list, final Comparator<? super Elements> comparator)
	{
		final int size = list.size();
		for (int i = 0; i < (size - 1); i++)
		{
			final int begin = i;
			final int end;// index of first element different from the ith element

			{// set end value
				int step = 1;

				while (comparator.compare(list.get(begin), list.get(begin + step)) == 0)
				{
					step = step << 1;
					if ((step + begin) >= size)
					{
						break;
					}
				}
				int lB = begin + (step >> 1);
				int uB = Math.min(begin + step, size);
				while (lB < (uB - 1))
				{
					final int testKey = lB + ((uB - lB) >> 1);
					if (comparator.compare(list.get(begin), list.get(testKey)) == 0)
					{
						lB = testKey;
					} else
					{
						uB = testKey;
					}

				}

				end = uB;
			}
			i = end - 1;

			Collections.shuffle(list.subList(begin, end));
		}
	}

	public static void shuffleWithinSmallRange(final List<?> finalDegreeIncreases)
	{
		ArrayS.shuffleWithinSmallRange(finalDegreeIncreases, ArrayS.PERCENTAGE_SHUFFLE_VERTEX_ORDER);
	}

	public static void shuffleWithinSmallRange(final List<?> finalDegreeIncreases, final float floatValue)

	{
		final int size = finalDegreeIncreases.size();
		if (size > 2)
		{
			final int range = Math.max(1, Math.round(size * floatValue));
			final int step = Math.max(range >> 1, 1);
			for (int startIndex = 0; startIndex < size; startIndex += step)
			{
				Collections.shuffle(finalDegreeIncreases.subList(startIndex, Math.min(size, startIndex + range)));
			}
		}
	}

	public static int[] sort(final int[] arr)
	{
		final int[] permutationIndices = new int[arr.length];
		for (int i = 0; i < permutationIndices.length; i++)
		{
			permutationIndices[i] = i;
		}

		final int off = 0;
		final int len = arr.length;
		ArrayS.sort1(arr, permutationIndices, off, len);
		return permutationIndices;
	}

	public static void sort1(final int[] arr, final int[] p, final int off, final int len)
	{
		// Insertion sort on smallest arrays
		if (len < 7)
		{
			for (int i = off; i < (len + off); i++)
			{
				for (int j = i; (j > off) && (arr[p[j - 1]] > arr[p[j]]); j--)
				{
					ArrayS.swap(p, j, j - 1);
				}
			}
			return;
		}

		// Choose a partition element, v
		int m = off + (len >> 1); // Small arrays, middle element
		if (len > 7)
		{
			int l = off;
			int n = (off + len) - 1;
			if (len > 40)
			{ // Big arrays, pseudomedian of 9
				final int s = len / 8;
				l = ArrayS.med3(arr, p, l, l + s, l + (2 * s));
				m = ArrayS.med3(arr, p, m - s, m, m + s);
				n = ArrayS.med3(arr, p, n - (2 * s), n - s, n);
			}
			m = ArrayS.med3(arr, p, l, m, n); // Mid-size, med of 3
		}
		final int v = arr[p[m]];

		// Establish Invariant: v* (<v)* (>v)* v*
		int a = off, b = a, c = (off + len) - 1, d = c;
		while (true)
		{
			while ((b <= c) && (arr[p[b]] <= v))
			{
				if (arr[p[b]] == v)
				{
					ArrayS.swap(p, a++, b);
				}
				b++;
			}
			while ((c >= b) && (arr[p[c]] >= v))
			{
				if (arr[p[c]] == v)
				{
					ArrayS.swap(p, c, d--);
				}
				c--;
			}
			if (b > c)
			{
				break;
			}
			ArrayS.swap(p, b++, c--);
		}

		// Swap partition elements back to middle
		int s;
		final int n = off + len;
		s = Math.min(a - off, b - a);
		ArrayS.vecswap(p, off, b - s, s);
		s = Math.min(d - c, n - d - 1);
		ArrayS.vecswap(p, b, n - s, s);

		// Recursively sort non-partition-elements
		if ((s = b - a) > 1)
		{
			ArrayS.sort1(arr, p, off, s);
		}
		if ((s = d - c) > 1)
		{
			ArrayS.sort1(arr, p, n - s, s);
		}
	}

	public static float standardDeviation(final float[] greedyTimeA)
	{
		final double mean = ArrayS.mean(greedyTimeA);
		double stb = 0;
		for (final double element : greedyTimeA)
		{
			stb += Math.pow(element - mean, 2);
		}
		stb /= greedyTimeA.length - 1;
		return (float) Math.sqrt(stb);
	}

	public static double standardDeviation(final int[] greedyK)
	{
		final double mean = ArrayS.mean(greedyK);
		double stb = 0;
		for (final int element : greedyK)
		{
			stb += Math.pow(element - mean, 2);
		}
		stb /= greedyK.length - 1;
		stb = Math.sqrt(stb);
		return stb;
	}

	/**
	 * Swaps x[a] with x[b].
	 */
	private static void swap(final int[] p, final int a, final int b)
	{
		final int t = p[a];
		p[a] = p[b];
		p[b] = t;
	}

	public static String toString(final byte[] arr, final String sep, final Format formater)
	{
		return ArrayS.toString(ArrayUtils.toObject(arr), sep, formater);
	}

	public static String toString(final double[] arr, final String sep, final Format formater)
	{
		return ArrayS.toString(ArrayUtils.toObject(arr), sep, formater);
	}

	public static String toString(final float[] arr, final String sep, final Format formater)
	{
		return ArrayS.toString(ArrayUtils.toObject(arr), sep, formater);
	}

	public static String toString(final int[] arr, final String sep, final Format formater)
	{
		return ArrayS.toString(ArrayUtils.toObject(arr), sep, formater);
	}

	public static String toString(final Iterable col, final String sep)
	{
		return ArrayS.toString(col, sep, NumberFormat.getInstance());
	}

	public static String toString(final Iterable col, final String sep, final Format formater)
	{
		final StringBuilder b = new StringBuilder();
		boolean first = true;
		for (final Object c : col)
		{
			b.append(first ? "" : sep);
			first = false;
			if (c instanceof int[])
			{
				b.append(Arrays.toString(ArrayUtils.toObject((int[]) c)));
			} else if (c instanceof float[])
			{
				b.append(Arrays.toString(ArrayUtils.toObject((float[]) c)));
			} else if (c instanceof Number)
			{
				b.append(formater.format(c));
			}
		}
		return b.toString();
	}

	public static String toString(final long[] arr, final String sep, final Format formater)
	{
		return ArrayS.toString(ArrayUtils.toObject(arr), sep, formater);
	}

	public static String toString(final Object[] col, final String sep)
	{
		return ArrayS.toString(col, sep, NumberFormat.getInstance());
	}

	public static String toString(final Object[] col, final String sep, final Format formater)
	{
		final StringBuilder b = new StringBuilder();
		boolean first = true;
		for (final Object c : col)
		{
			b.append(first ? "" : sep);
			first = false;
			if (c instanceof Number)
			{
				b.append(formater.format(c));
			} else
			{
				b.append(c);
			}
		}
		return b.toString();
	}

	/**
	 * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
	 */
	private static void vecswap(final int p[], int a, int b, final int n)
	{
		for (int i = 0; i < n; i++, a++, b++)
		{
			ArrayS.swap(p, a, b);
		}
	}

}
