package distanceFunction;

import java.util.Vector;

public abstract class SymmetricBooleanDistanceFunction
{

	private final int noElements;

	public SymmetricBooleanDistanceFunction(final int n)
	{
		this.noElements = n;
	}

	protected void checkIndices(final int i, final int j) throws IllegalAccessError
	{
		if ((i >= this.getNoOfElements()) || (j >= this.getNoOfElements()) || (i == j))
			throw new IllegalAccessError(i + "," + j);
	}

	/**
	 * Returns d(i,j).
	 * 
	 * @param i index
	 * @param j index
	 * @return
	 */
	public boolean get(final int i, final int j)
	{
		this.checkIndices(i, j);

		if (j < i)
			return this.getInternal(j, i - (j + 1));
		else
			return this.getInternal(i, j - (i + 1));
	}

	abstract boolean getInternal(final int i, final int j);

	public int getNoOfElements()
	{
		return this.noElements;
	}

	public void set(final int i, final int j, final boolean value)
	{
		this.checkIndices(i, j);

		if (j < i)
		{
			this.setInternal(j, i - (j + 1), value);
		} else
		{
			this.setInternal(i, j - (i + 1), value);
		}
	}

	abstract void setInternal(final int i, final int j, boolean value);

	@Override
	public String toString()
	{
		final StringBuffer s = new StringBuffer();
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			s.append(i + "\t");
		}
		s.append('\n');
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			s.append("_______");
		}
		s.append('\n');
		for (int i = 0; i < this.getNoOfElements(); i++)
		{
			final Vector<Boolean> vec = new Vector<Boolean>();
			for (int j = i + 1; j < this.getNoOfElements(); j++)
			{
				vec.add(this.get(i, j));
			}

			// final Boolean[] arr = vec.size() > 0 ? (Boolean[])
			// vec.toArray(new Boolean[vec.size()]) : new Boolean[] {};
			s.append(i + "|\t");
			for (int t = 0; t < i; t++)
			{
				s.append("\t");
			}
			for (int j = 0; j < vec.size(); j++)
			{
				if (j == 0)
				{
					s.append("{");
				}
				s.append(vec.get(j) ? "1" : "0");

				if (j < (vec.size() - 1))
				{
					s.append(",");
					s.append("\t");
				} else
				{
					s.append("},");
				}
			}
			s.append('\n');
		}
		s.append('\n');
		return s.toString();
	}
}
