package distanceFunction;

import java.util.Arrays;

public class SymmetricDistanceFunctionIntFunction<Value>
{
	private final Value[][] distance;

	private final int noElements;

	public SymmetricDistanceFunctionIntFunction(final int n)
	{
		this.distance = (Value[][]) new Object[n - 1][];
		this.noElements = this.distance.length + 1;
		for (int i = n - 1; i > 0; i--)
		{
			this.distance[n - 1 - i] = (Value[]) new Object[i];

		}
	}

	private SymmetricDistanceFunctionIntFunction(final Value[][] arr)
	{
		this.distance = arr;
		this.noElements = this.distance.length + 1;
	}

	@Override
	public SymmetricDistanceFunctionIntFunction<Value> clone()
	{
		final Value[][] copyArr = (Value[][]) new Object[this.distance.length][];
		for (int i = 0; i < copyArr.length; i++)
		{
			copyArr[i] = this.distance[i].clone();
		}
		return new SymmetricDistanceFunctionIntFunction<Value>(copyArr);
	}

	/**
	 * Returns d(i,j).
	 * 
	 * @param i index
	 * @param j index
	 * @return
	 */
	public Value get(final int i, final int j)
	{
		if (j < i)
			return this.distance[j][i - (j + 1)];
		else
			return this.distance[i][j - (i + 1)];
	}

	public int getNoOfElements()
	{
		return this.noElements;
	}

	@Override
	public int hashCode()
	{
		return Arrays.deepHashCode(this.distance) * this.noElements;
	}

	public void set(final int i, final int j, final Value value)
	{
		if (j < i)
		{
			this.distance[j][i - (j + 1)] = value;
		} else
		{
			this.distance[i][j - (i + 1)] = value;
		}
	}

	@Override
	public String toString()
	{
		final StringBuffer s = new StringBuffer();
		for (int i = 0; i <= this.distance.length; i++)
		{
			s.append(i + "\t");
		}
		s.append('\n');
		for (int i = 0; i <= this.distance.length; i++)
		{
			s.append("_______");
		}
		s.append('\n');
		for (int i = 0; i < this.distance.length; i++)
		{
			final Value[] arr = this.distance[i];
			s.append(i + "|\t");
			for (int t = 0; t < i; t++)
			{
				s.append("\t");
			}
			for (int j = 0; j < arr.length; j++)
			{
				if (j == 0)
				{
					s.append("{");
				}
				s.append(arr[j]);

				if (j < (arr.length - 1))
				{
					s.append(",");
					s.append("\t");
				} else
				{
					s.append("},");
				}
			}
			s.append('\n');
		}
		s.append('\n');
		return s.toString();
	}
}
