package ProgOpt;

import java.io.IOException;
import java.util.ArrayList;

import org.apache.commons.lang3.mutable.MutableBoolean;
import org.apache.commons.lang3.mutable.MutableDouble;
import org.apache.commons.lang3.mutable.MutableInt;

import solver.heuristics.GreedyLowerBoundHeuristic;
import solver.heuristics.GreedyUpperBoundHeuristic;
import solver.hierarchicalClustering.SearchTreeSolver;
import solver.lpBased.HierClusterLpSolver;
import benchmark.helpClasses.BenchResult;
import benchmark.helpClasses.TimeBoundedApplication;
import benchmark.helpClasses.timeBoundedApplications.FindOptimalSolutionWithinTime;

import commandLineOptionsHelpClasses.BooleanOpt;
import commandLineOptionsHelpClasses.CLOption;
import commandLineOptionsHelpClasses.DoubleOpt;
import commandLineOptionsHelpClasses.IntOpt;
import commandLineOptionsHelpClasses.StringOpt;

import dataReductionRules.Rule1;
import dataStructures.Store;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public class SmacClusterEditOpt
{
	/**
	 * Max. number of running time seconds.
	 */
	public final static MutableDouble cutoff_time = new MutableDouble(300d);

	public static int getOptimalK(final String fileName, final boolean interleaveDR, final boolean interleaveLP) throws IOException
	{
		final Store<SymmetricDiscreteDistanceFunction, Store<ArrayList<int[]>, ArrayList<int[]>>> store = SmacClusterEditOpt
				.readInstanceFile(fileName);
		final SymmetricDiscreteDistanceFunction dist = store.getS1();
		final Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks = store.getS2();

		final long startTime = System.currentTimeMillis();

		final FindOptimalSolutionWithinTime d = new FindOptimalSolutionWithinTime((int) SmacClusterEditOpt.cutoff_time.getValue().doubleValue(),
				dist, modBlocks, interleaveDR, true, interleaveLP);
		// d.runAlgorithm();
		d.run();
		final BenchResult result = d.getBenchResult();

		if (result.getTime() == TimeBoundedApplication.STOP_TIME_VALUE)
		{
			final double e = (System.currentTimeMillis() - startTime) / 1000d;
			final String s = "Result for SMAC:" + " TIMEOUT, " + e + "," + e + "," + -1 + ", 0";
			System.out.println(s);
			return Integer.MAX_VALUE;
		} else
		{
			System.out.println("opt k:" + result.getParameter());
			final String s = "Result for SMAC:" + " SUCCESS, " + result.getTime() + ", " + result.getTime() + ", " + result.getParameter() + ", 0";
			System.out.println(s);
			return (int) result.getParameter();
		}

	}

	/**
	 * @param args
	 * @throws IOException
	 */
	public static void main(final String[] args) throws IOException
	{
		// SearchTreeSolver.RulesVerboseLevel = 1;
		// SearchTreeSolver.SearchTreeSolver_LogLevel = 1;

		final MutableBoolean verbose = new MutableBoolean(true);
		new BooleanOpt(verbose, "verbose", false, false).setValue(args);

		final StringBuilder fileNameB = new StringBuilder();

		final MutableBoolean interleaveDR = new MutableBoolean(true);
		final MutableBoolean interleaveLP = new MutableBoolean(true);
		final MutableDouble interleaveLP_BoundFraction_Small = new MutableDouble(0.4d);

		SmacClusterEditOpt.setAndReadCommandLineParameters(args, verbose, fileNameB, interleaveDR, interleaveLP, interleaveLP_BoundFraction_Small);
		{
			SearchTreeSolver.interleaveLP_Bound_Small = Math.max(1,
					Math.round(interleaveLP_BoundFraction_Small.floatValue() * SearchTreeSolver.interleaveLP_Bound_Medium.intValue()));
		}
		System.out.println("lpSmallConstant|lpMediumConstant|lpLargeConstant>lpSmallBound|lpMediumBound");
		System.out.println(SearchTreeSolver.interleaveLP_Constant_Small + "|" + SearchTreeSolver.interleaveLP_Constant_Medium + "|"
				+ SearchTreeSolver.interleaveLP_Constant_Large + ">" + SearchTreeSolver.interleaveLP_Bound_Small + "|"
				+ SearchTreeSolver.interleaveLP_Bound_Medium);
		System.out.println("interleaveDR_Small_Constant|interleaveDR_Large_Constant>interleaveDR_Large_Bound");
		System.out.println(SearchTreeSolver.interleaveDR_Constant_Small + "|" + SearchTreeSolver.interleaveDR_Constant_Large + ">"
				+ SearchTreeSolver.interleaveDR_Bound_Large);

		SmacClusterEditOpt.getOptimalK(fileNameB.toString(), interleaveDR.getValue(), interleaveLP.getValue());
		System.out.flush();
	}

	public static Store<SymmetricDiscreteDistanceFunction, Store<ArrayList<int[]>, ArrayList<int[]>>> readInstanceFile(final String fileName)
			throws IOException
	{
		final SymmetricDiscreteDistanceFunction dist;
		Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks = null;
		if (fileName.endsWith("dat"))
		{
			dist = SymmetricDiscreteDistanceFunction.createInstanceFromFile(fileName);
		} else if (fileName.endsWith("graph"))
		{
			dist = SymmetricDiscreteDistanceFunction.createInstanceFromGraphFile(fileName);
		} else
		// cm file
		{
			final Store<Store<SymmetricDiscreteDistanceFunction, Integer>, Store<ArrayList<int[]>, ArrayList<int[]>>> distanceFunctionPlusBlocks = SymmetricDiscreteDistanceFunction
					.createSymmetricDiscreteDistanceFunctionInstanceFromCostMatrixFile(fileName);
			System.out.println("totalReverseDRCorrect:" + distanceFunctionPlusBlocks.getS1().getS2());
			dist = distanceFunctionPlusBlocks.getS1().getS1();
			modBlocks = distanceFunctionPlusBlocks.getS2();
		}
		return new Store<SymmetricDiscreteDistanceFunction, Store<ArrayList<int[]>, ArrayList<int[]>>>(dist, modBlocks);
	}

	public static void setAndReadCommandLineParameters(final String[] args, final MutableBoolean verbose, final StringBuilder fileNameB,
			final MutableBoolean interleaveDR, final MutableBoolean interleaveLP, final MutableDouble interleaveLP_BoundFraction_Small)
	{
		final ArrayList<CLOption> commandLineOptions = new ArrayList<CLOption>();
		{// read commandline arguments

			commandLineOptions.add(new StringOpt(fileNameB, "instance_name", 0, verbose.getValue(), true));
			commandLineOptions.add(new StringOpt(new StringBuilder(), "instance_specific infor", 1, verbose.getValue(), true));
			commandLineOptions.add(new DoubleOpt(SmacClusterEditOpt.cutoff_time, "cutoff_time", 2, verbose.getValue(), true));
			commandLineOptions.add(new DoubleOpt(new MutableDouble(), "cutoff_length", 3, verbose.getValue(), true));
			// commandLineOptions.add(new DoubleOpt(new MutableDouble(), "seed",
			// 4, verbose.getValue(), true));
		}
		{// increase steps for FindOptimalSolutionsWithinTime
			commandLineOptions.add(new DoubleOpt(FindOptimalSolutionWithinTime.stepIncreaseExponent, "StepIncreaser_StepIncreaseExponent", verbose
					.getValue(), false));
			commandLineOptions.add(new BooleanOpt(FindOptimalSolutionWithinTime.alwaysRoundStepsToPowOfTwo,
					"StepIncreaser_AlwaysRoundStepsToPowOfTwo", verbose.getValue(), false));
		}
		{// Greedyheuristic

			// no longer officially supported; but still works
			commandLineOptions.add(new IntOpt(GreedyLowerBoundHeuristic.heuristicLowerBoundRuns, "GreedyheuristicLowerBoundRuns", verbose.getValue(),
					false));

			commandLineOptions.add(new DoubleOpt(GreedyLowerBoundHeuristic.lowerBoundTimeProportion, "lowerBoundTimeProportion", verbose.getValue(),
					false));
			commandLineOptions.add(new DoubleOpt(GreedyLowerBoundHeuristic.highestShuffleFrac, "GreedyLowerBound_highestShuffleFrac", verbose
					.getValue(), false));
			commandLineOptions.add(new IntOpt(GreedyLowerBoundHeuristic.GreedyLowerBound_modShuffle, "GreedyLowerBound_modShuffle", verbose
					.getValue(), false));
			commandLineOptions.add(new DoubleOpt(FindOptimalSolutionWithinTime.GreedyLowerBound_Time_Frac, "GreedyLowerBound_Time_Frac", verbose
					.getValue(), false));

			commandLineOptions.add(new IntOpt(FindOptimalSolutionWithinTime.numOfIndpendentGreedyRuns, "GreedyLowerBound_numOfIndpendentGreedyRuns",
					verbose.getValue(), false));

			commandLineOptions.add(new IntOpt(GreedyUpperBoundHeuristic.GreedyheuristicUpperBoundRuns, "GreedyheuristicUpperBoundRuns", verbose
					.getValue(), false));

			commandLineOptions.add(new DoubleOpt(GreedyUpperBoundHeuristic.upperBoundTimeProportion, "upperBoundTimeProportion", verbose.getValue(),
					false));
			
			commandLineOptions.add(new DoubleOpt(GreedyUpperBoundHeuristic.RANDOM_ADD_COMPARISON, "GreedyUpperBound_RANDOM_ADD_COMPARISON", verbose
					.getValue(), false));
			commandLineOptions.add(new DoubleOpt(GreedyUpperBoundHeuristic.RANDOM_SCORE_ADD, "GreedyUpperBound_RANDOM_SCORE_ADD", verbose
					.getValue(), false));
			commandLineOptions.add(new IntOpt(GreedyUpperBoundHeuristic.Max_Queue_Size, "GreedyUpperBound_Max_Queue_Size",
					verbose.getValue(), false));
		}

		{// parameters for LP lower bound
			commandLineOptions.add(new IntOpt(FindOptimalSolutionWithinTime.LP_LowerBound_N_Threshold, "LP_LowerBound_N_Threshold", verbose
					.getValue(), false));

			commandLineOptions.add(new IntOpt(HierClusterLpSolver.LP_LowerBound_ILP_Part_First_Run, "LP_LowerBound_ILP_Part_First_Run", verbose
					.getValue(), false));

			commandLineOptions.add(new DoubleOpt(FindOptimalSolutionWithinTime.LP_LowerBound_Time_Fac, "LP_LowerBound_Time_Fac", verbose.getValue(),
					false));
		}

		{// just for testing, nothing to optimize
			commandLineOptions.add(new BooleanOpt(SymmetricDiscreteDistanceFunction.writerUncompressedCM_File, "writerUncompressedCM_File", verbose
					.getValue(), false));
			commandLineOptions.add(new IntOpt(SymmetricDiscreteDistanceFunction.verbosityLevelWhenReadingCM, "verbosityLevelWhenReadingCM", verbose
					.getValue(), false));

		}

		commandLineOptions.add(new BooleanOpt(interleaveDR, "interleaveDR", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveDR_Bound_Large, "interleaveDR_Bound_Large", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveDR_Constant_Large, "interleaveDR_Constant_Large", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveDR_Constant_Small, "interleaveDR_Constant_Small", verbose.getValue(), false));

		// no longer active
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule0, "rule0", verbose.getValue(), false));

		// no longer active
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule1, "rule1", verbose.getValue(), false));

		commandLineOptions.add(new IntOpt(Rule1.queueSize, "rule1QueueSize", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(Rule1.useSumAsCount, "rule1UseSumAsCount", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(Rule1.maxUnsuccess, "rule1MaxUnsuccess", verbose.getValue(), false));

		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule2, "rule2", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule3, "rule3", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule4, "rule4", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule5, "rule5", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule6, "rule6", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.Apply_Rule7, "rule7", verbose.getValue(), false));

		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule0_InterleavingConst, "Rule0_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule1_InterleavingConst, "Rule1_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule2_InterleavingConst, "Rule2_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule3_InterleavingConst, "Rule3_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule4_InterleavingConst, "Rule4_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule5_InterleavingConst, "Rule5_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule6_InterleavingConst, "Rule6_InterleavingConst", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.Rule7_InterleavingConst, "Rule7_InterleavingConst", verbose.getValue(), false));

		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.randomizeRuleOrderAtBeginning, "randomizeRuleOrderAtBeginning", verbose.getValue(),
				false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.randomizeRuleOrderEachApplication, "randomizeRuleOrderEachApplication", verbose
				.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.applyRulesExhaustively, "applyRulesExhaustively", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.offset_DR, "offset_DR", verbose.getValue(), false));
		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.useLevelForKernelInterleavingConstant, "useLevelForKernelInterleavingConstant",
				verbose.getValue(), false));

		commandLineOptions.add(new BooleanOpt(SearchTreeSolver.useLevelForLPInterleavingConstant, "useLevelForLPInterleavingConstant", verbose
				.getValue(), false));
		commandLineOptions.add(new BooleanOpt(interleaveLP, "interleaveLP", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveLP_Bound_Medium, "interleaveLP_Bound_Medium", verbose.getValue(), false));
		commandLineOptions.add(new DoubleOpt(interleaveLP_BoundFraction_Small, "interleaveLP_BoundFraction_Small", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveLP_Constant_Large, "interleaveLP_Constant_Large", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveLP_Constant_Medium, "interleaveLP_Constant_Medium", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.interleaveLP_Constant_Small, "interleaveLP_Constant_Small", verbose.getValue(), false));
		commandLineOptions.add(new IntOpt(SearchTreeSolver.offset_LP, "offset_LP", verbose.getValue(), false));

		for (final CLOption k : commandLineOptions)
		{
			k.setValue(args);
		}
	}

}
