package benchmark.helpClasses;

import java.util.Iterator;
import java.util.TreeMap;

public class BenchResult
{
	private static final int ITER_DEFAULT = -1;

	private final TreeMap<String, Object> addParameters = new TreeMap<String, Object>();

	private long iter;

	private float lpTimeSec = -1f;

	private float lpVal;

	private int parCount;

	private double parSum;

	private int timeCount;

	private double timeSum;

	public BenchResult()
	{
		// TODO Auto-generated constructor stub
	}

	public BenchResult(final BenchResult[] results)
	{
		for (final BenchResult r : results)
		{
			if (r != null)
			{
				this.add(r);
			}
		}
	}

	public BenchResult(final double param, final double timeInSec)
	{
		this(param, timeInSec, BenchResult.ITER_DEFAULT);
	}

	public BenchResult(final double parameter, final double timeInSeconds, final long iter)
	{
		this(parameter, timeInSeconds, iter, BenchResult.ITER_DEFAULT, BenchResult.ITER_DEFAULT);
	}

	public BenchResult(final double parameter, final double timeInSeconds, final long iter, final float lpVal, final float lpTime)
	{
		if (parameter != Integer.MAX_VALUE)
		{
			this.parSum = parameter;
			this.parCount = 1;
		}
		if (timeInSeconds != TimeBoundedApplication.STOP_TIME_VALUE)
		{
			this.timeCount = 1;
			this.timeSum = timeInSeconds;
		}
		this.iter = iter;
		this.lpVal = lpVal;
		this.lpTimeSec = lpTime;
	}

	public void add(final BenchResult r)
	{
		if (r != null)
		{
			if (r.getTime() != TimeBoundedApplication.STOP_TIME_VALUE)
			{
				this.timeCount++;
				this.timeSum += r.getTime();
			}
			if ((r.getParameter() != Integer.MAX_VALUE) && (r.getParameter() != TimeBoundedApplication.STOP_TIME_VALUE))
			{
				this.parSum += r.getParameter();
				this.parCount++;
			}
			this.iter = +r.iter;

			for (final Iterator<String> iterator = r.addParameters.keySet().iterator(); iterator.hasNext();)
			{
				final String key = iterator.next();
				if (this.addParameters.containsKey(key))
				{
					if ((this.addParameters.get(key) instanceof String) && (r.addParameters.get(key) instanceof String))
					{
						this.addParameters.put(key, (String) this.addParameters.get(key) + (String) r.addParameters.get(key));
					}
					if ((this.addParameters.get(key) instanceof Integer) && (r.addParameters.get(key) instanceof Integer))
					{
						this.addParameters.put(key, (Integer) this.addParameters.get(key) + (Integer) r.addParameters.get(key));
					}
				} else
				{
					this.addParameters.put(key, r.addParameters.get(key));
				}
			}
		}

	}

	public float getLpTimeInSec()
	{
		return this.lpTimeSec;
	}

	/**
	 * Return the lower bound, found by the lp.
	 * 
	 * @return
	 */
	public float getLPVal()
	{
		return this.lpVal;
	}

	public long getNumberRecursiveSteps()
	{
		return this.iter;
	}

	public double getParameter()
	{
		return this.parCount > 0 ? this.parSum / this.parCount : TimeBoundedApplication.STOP_TIME_VALUE;
	}

	public String getParameter(final String key)
	{
		if (this.addParameters.containsKey(key))
			return this.addParameters.get(key).toString();
		return null;
	}

	public double getTime()
	{
		return this.timeCount > 0 ? this.timeSum / this.timeCount : TimeBoundedApplication.STOP_TIME_VALUE;
	}

	public void setLPTime(final long l)
	{
		this.lpTimeSec = (float) (l / 1000d);
	}

	public void setLPVal(final float lp)
	{
		this.lpVal = lp;
	}

	public void setParameter(final int optK)
	{
		this.parSum = optK;
		this.parCount = 1;
	}

	public void setParameter(final String key, final Object ob)
	{
		this.addParameters.put(key, ob);
	}

	@Override
	public String toString()
	{
		return String.valueOf(this.getTime());
	}
}