package benchmark.helpClasses.timeBoundedApplications;

import java.util.ArrayList;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.commons.lang3.mutable.MutableBoolean;
import org.apache.commons.lang3.mutable.MutableDouble;
import org.apache.commons.lang3.mutable.MutableInt;

import solver.heuristics.GreedyLowerBoundHeuristic;
import solver.heuristics.GreedyUpperBoundHeuristic;
import solver.hierarchicalClustering.Conflicts;
import solver.hierarchicalClustering.SearchTreeSolver;
import solver.lpBased.FoundILPSolution;
import solver.lpBased.HierClusterLpSolver;
import ProgOpt.SmacClusterEditOpt;
import benchmark.helpClasses.BenchResult;
import benchmark.helpClasses.StopWatch;
import benchmark.helpClasses.TimeBoundedApplication;
import dataStructures.Store;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public class FindOptimalSolutionWithinTime extends TimeBoundedApplication
{
	class GreedyThread implements Callable<Integer>
	{
		private final GreedyLowerBoundHeuristic greedyLowerBoundHeuristic;
		private final float highShuffle;
		private final double maxTime;
		private final float minShuffle;
		private final int numberOfRounds;

		GreedyThread(final GreedyLowerBoundHeuristic greedyLowerBoundHeuristic, final float highShuffle, final float minShuffle, final int numRounds,
				final double maxTime)
		{
			this.greedyLowerBoundHeuristic = greedyLowerBoundHeuristic;
			this.highShuffle = highShuffle;
			this.minShuffle = minShuffle;
			this.numberOfRounds = numRounds;
			this.maxTime = maxTime;
		}

		@Override
		public Integer call() throws Exception
		{
			final StopWatch watch = new StopWatch();

			final int greedy_LowerBound = this.greedyLowerBoundHeuristic.getLowerBound(this.maxTime, this.highShuffle, this.minShuffle,
					this.numberOfRounds);

			watch.printTime("greedyLB:" + greedy_LowerBound + " ", System.out);
			System.out.println();
			return greedy_LowerBound;
		}
	}

	class LPThread implements Callable<Store<Integer, Boolean>>
	{
		private final SymmetricDiscreteDistanceFunction dist;
		private final double maxTimeInSeconds;

		LPThread(final SymmetricDiscreteDistanceFunction d, final double maxTimeSec)
		{
			this.dist = d;
			this.maxTimeInSeconds = FindOptimalSolutionWithinTime.LP_LowerBound_Time_Fac.doubleValue() * maxTimeSec;
		}

		@Override
		public Store<Integer, Boolean> call() throws Exception
		{
			final StopWatch watch = new StopWatch();
			try
			{
				final int lB = this.dist.getNoOfElements() < FindOptimalSolutionWithinTime.LP_LowerBound_N_Threshold.intValue() ? HierClusterLpSolver
						.getLpLowerBound(this.dist, this.maxTimeInSeconds, -1d) : 0;
				watch.printTime("LP lowerbound:" + lB + " ", System.out);
				System.out.println();

				return new Store<Integer, Boolean>(lB, false);
			} catch (final FoundILPSolution e)
			{
				watch.printTime("LP lowerbound:" + e.solutionSize + "(opt) ", System.out);
				System.out.println();

				return new Store<Integer, Boolean>(e.solutionSize, true);
			}

		}

	}

	public static MutableBoolean alwaysRoundStepsToPowOfTwo = new MutableBoolean(false);

	public static final MutableDouble GreedyLowerBound_Time_Frac = new MutableDouble(0.75);

	public final static MutableInt LP_LowerBound_N_Threshold = new MutableInt(100);

	// private final int beginK;

	public static final MutableDouble LP_LowerBound_Time_Fac = new MutableDouble(2);

	public final static MutableInt numOfIndpendentGreedyRuns = new MutableInt(2);

	public static final MutableDouble stepIncreaseExponent = new MutableDouble(1.2);

	public static double log_B2(final int l)
	{
		return Math.log(l) / Math.log(2);
	}

	/**
	 * @param greedyResult
	 * @return
	 */
	private static int readGreedyResult(final Future<Integer> greedyResult)
	{
		final int greedy_LowerBound;
		{// read greedy heuristic result

			Integer threadResult = null;
			try
			{
				threadResult = greedyResult.get();
			} catch (InterruptedException | ExecutionException e1)
			{
				e1.printStackTrace();
			}
			// System.err.println("greed null:" + (threadResult == null));
			greedy_LowerBound = threadResult != null ? threadResult : 0;
		}
		return greedy_LowerBound;
	}

	private final SymmetricDiscreteDistanceFunction dist;

	private final boolean INTERLEAVE_LP;

	public final boolean INTERLEAVING;

	public final boolean KERNELIZATION;

	private final Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks;

	private int optK;

	private BenchResult result = new BenchResult(Integer.MAX_VALUE, TimeBoundedApplication.STOP_TIME_VALUE, -1);

	public FindOptimalSolutionWithinTime(final int sec, final SymmetricDiscreteDistanceFunction dist, final boolean interleaveDR,
			final boolean interleaveLP, final boolean kernelization)
	{
		this(sec, dist, null, interleaveDR, interleaveLP, kernelization);
	}

	public FindOptimalSolutionWithinTime(final int sec, final SymmetricDiscreteDistanceFunction dist,
			final Store<ArrayList<int[]>, ArrayList<int[]>> modBlocks, final boolean interleaveDR, final boolean interleaveLP,
			final boolean kernelization)
	{
		super(sec);
		this.dist = dist;
		this.modBlocks = modBlocks;
		this.INTERLEAVE_LP = interleaveLP;
		this.INTERLEAVING = interleaveDR;
		this.KERNELIZATION = kernelization;
	};

	public BenchResult getBenchResult()
	{
		if (this.result.getParameter() == TimeBoundedApplication.STOP_TIME_VALUE)
		{
			this.result.setParameter(this.optK);
		}
		return this.result;
	}

	// create instance to test
	// /home/sepphar/localRepos/progOptClustEdit/instancesHierClust/realWorld/c_66_M_2/nr_1651_n_30_c_0.33_M_2.dat
	// /home/sepphar/localRepos/progOptClustEdit/instances_peace/reduced_samples_V_100to1000_k_0.5to10V/reduced_samples_V_50to1000_k_0.5to10V_10samples_set1/random_graphs_reduced_V_150-199_k_6V/graph_Vred_150_krest_930_c_1050_Vorg_150_korg_930_lb_888_ub_973_nr_0.cm
	private int getParameterIncreaseStep(final int run, final int param, final double lastTime)
	{
		final int numberOfRemainingRuns = (int) Math.floor((this.maxNumberMillis - (System.currentTimeMillis() - this.beginTime)) / 1000d / lastTime);
		int s;

		if (run == 0)
		{
			s = 4;
		} else
		{
			s = (int) Math.pow(run + 3, FindOptimalSolutionWithinTime.stepIncreaseExponent.doubleValue());
		}
		if (FindOptimalSolutionWithinTime.alwaysRoundStepsToPowOfTwo.booleanValue())
		{
			s = 1 << (int) Math.ceil(FindOptimalSolutionWithinTime.log_B2(s));
		}

		final int out = Math.min(1 << Math.min(30, Math.max(0, numberOfRemainingRuns - 1)), s);
		// System.err.println("\nreturn:" + out + "(s:" + s + ",run:" + run +
		// ",num:" + numberOfRemainingRuns + ")");
		return out;
		//
		// //with x remaining runs, max return can be at most 2^(x-1)
		//
		// // (x is at least 3+4)
		// int x = Math.min(run + 1, numberOfRemainingRuns - 1) + 4;
		// // 1.72^x , result at least 5
		// // int s = (int) Math.pow(1.72, x);
		// int s = (int) Math.round(Math.pow(x, 1.2));// x * (x - 1);
		// System.out.println("return:" + s);
		// return s;

		// {
		// // System.err.println("run:" + 1);
		// return 1;
		// }
		// int step = (int) Math.pow(1.5, Math.min(run + 1, Math.max(0,
		// numberOfRemainingRuns - 1)));
		// System.err.println("run:" + step);
		// return step;
	}

	private int getParameterIncreaseStep2(final int run, final int param, final double lastTime)
	{
		final double remainTime = (this.maxNumberMillis - (System.currentTimeMillis() - this.beginTime)) / 1000d;

		if (lastTime <= (0.01 * remainTime))
			return 5;
		else if (lastTime <= (0.1 * remainTime))
			return 4;
		else if (lastTime <= (0.2 * remainTime))
			return 3;
		else if (lastTime <= (0.3 * remainTime))
			return 2;
		else
			return 1;
	}

	private Store<Integer, Integer> readLPResult(final Future<Store<Integer, Boolean>> lpResult)
	{
		// read lp lowerbound result
		Store<Integer, Boolean> threadResult = null;
		try
		{
			threadResult = lpResult.get();
		} catch (InterruptedException | ExecutionException e)
		{
			e.printStackTrace();
		}
		// System.err.println("lp null:" + (threadResult == null));
		final int lp_lowerbound = threadResult != null ? threadResult.getS1() : 0;
		int upperBound = Integer.MAX_VALUE;
		if ((threadResult != null) && threadResult.getS2())
		{// lp result ilp and thus optimal?
			upperBound = threadResult.getS1();
		}
		return new Store<Integer, Integer>(lp_lowerbound, upperBound);
	}

	@Override
	public void run()
	{
		final int numberOfIndpendentGreedyRuns = FindOptimalSolutionWithinTime.numOfIndpendentGreedyRuns.intValue();
		this.beginTime = System.currentTimeMillis();
		final long begin = System.currentTimeMillis();
		long iter = 0;

		double lastStepTime = 0;

		// if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
		{
			System.out.println("start computing optimal k (n=" + this.dist.getNoOfElements() + ")");
		}

		final int lowerBound;
		int upperBound = Integer.MAX_VALUE;
		final Conflicts conflicts;

		{// compute lower bound on the solution size (lp and heuristic will be run in parallel)
			final ExecutorService executor = Executors.newCachedThreadPool();
			final double maxTime = (GreedyLowerBoundHeuristic.lowerBoundTimeProportion.doubleValue() * SmacClusterEditOpt.cutoff_time.doubleValue());
			final double maxTimeGreedy = maxTime * FindOptimalSolutionWithinTime.GreedyLowerBound_Time_Frac.doubleValue();

			final ArrayList<Future<Integer>> greedyResults = new ArrayList<>();// , greedyResult2, greedyResult3;
			final Future<Store<Integer, Boolean>> lpResult;

			// submit lp lowerbound runs
			lpResult = executor.submit(new LPThread(this.dist, maxTime));

			{// submit greedy lowerbound runs
				final GreedyLowerBoundHeuristic greedyLowerBoundHeuristic = new GreedyLowerBoundHeuristic(this.dist);
				conflicts = greedyLowerBoundHeuristic.getConflictStorage().clone();

				final float highShuffle = GreedyLowerBoundHeuristic.highestShuffleFrac.floatValue();

				greedyResults.add(executor.submit(new GreedyThread(greedyLowerBoundHeuristic, highShuffle, highShuffle
						* ((numberOfIndpendentGreedyRuns - 1) / (float) numberOfIndpendentGreedyRuns),
						GreedyLowerBoundHeuristic.heuristicLowerBoundRuns.intValue() / numberOfIndpendentGreedyRuns, maxTimeGreedy)));

				for (int gr = 2; gr <= numberOfIndpendentGreedyRuns; gr++)
				{
					executor.submit(new GreedyThread(new GreedyLowerBoundHeuristic(this.dist, conflicts.clone()), highShuffle
							* (((numberOfIndpendentGreedyRuns - gr) + 1) / (float) numberOfIndpendentGreedyRuns), highShuffle
							* ((numberOfIndpendentGreedyRuns - gr) / (float) numberOfIndpendentGreedyRuns),
							GreedyLowerBoundHeuristic.heuristicLowerBoundRuns.intValue() / numberOfIndpendentGreedyRuns, maxTimeGreedy));
				}
			}

			executor.shutdown();

			waitUntilFirstResultIsAvailable(lpResult, greedyResults);

			int lp_lowerbound = -1;
			if (lpResult.isDone())
			{
				// System.err.println("lp done");
				final Store<Integer, Integer> s = this.readLPResult(lpResult);
				lp_lowerbound = Math.max(s.getS1(), 0);
				upperBound = s.getS2();
			}

			if (upperBound == Integer.MAX_VALUE)// can be false only if LP was optimal
				upperBound = GreedyUpperBoundHeuristic.computeUpperBound(this.dist, conflicts);

			if (lp_lowerbound < 0)// not yet read
			{
				final Store<Integer, Integer> s = this.readLPResult(lpResult);
				lp_lowerbound = Math.max(s.getS1(), 0);
				upperBound = Math.min(upperBound, s.getS2());
			}

			int greedy_LowerBound = 0;
			for (final Future<Integer> greedyResult : greedyResults)
			{
				greedy_LowerBound = Math.max(greedy_LowerBound, FindOptimalSolutionWithinTime.readGreedyResult(greedyResult));
			}

			lowerBound = Math.max(greedy_LowerBound, lp_lowerbound);

			executor.shutdownNow();

			System.out.println("lowerbound: " + lowerBound);
			System.out.println("upperbound: " + upperBound);
			System.out.println(((System.currentTimeMillis() - this.beginTime) / 1000d) + "s");
		}

		if (lowerBound > upperBound)
			throw new RuntimeException("lowerbound larger than upperbound!!!!");

		if (lowerBound < upperBound)
		{
			int lastUnsolK = lowerBound - 1;
			int run = 0;
			final int maxStep = Math.max(1, upperBound != Integer.MAX_VALUE ? (upperBound - lowerBound) / 3 : 1);
			for (this.optK = lowerBound;; this.optK += Math.min(upperBound - this.optK,
					Math.max(maxStep, this.getParameterIncreaseStep(run++, this.optK, lastStepTime))))
			{
				lastStepTime = System.currentTimeMillis();

				SearchTreeSolver solver = new SearchTreeSolver(conflicts.clone(), this.dist, this.modBlocks, this.optK, false);
				final boolean solverSuccess = solver.solve(this.KERNELIZATION, this.INTERLEAVING, this.INTERLEAVE_LP);

				if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
				{
					System.out.println("test:" + this.optK + "(" + lastUnsolK + ",time:" + ((System.currentTimeMillis() - lastStepTime) / 1000d)
							+ "s)");
				}

				if (solverSuccess)
				{
					if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
					{
						System.out.println("solved:" + this.optK + "(" + lastUnsolK + ")");
					}
					iter = solver.getNumberOfRecursiveSteps();

					// binary search for right param
					while ((lastUnsolK + 1) < this.optK)
					{
						lastStepTime = System.currentTimeMillis();// only for
																	// logging
						final int mid = lastUnsolK + Math.max(1, (this.optK - lastUnsolK) / 2);

						final SearchTreeSolver solvertmp = new SearchTreeSolver(this.dist, this.modBlocks, mid);
						final boolean solve = solvertmp.solve(this.KERNELIZATION, this.INTERLEAVING, this.INTERLEAVE_LP);

						if (SearchTreeSolver.SearchTreeSolver_LogLevel > 0)
						{
							System.out.print("binary:" + mid + "(" + this.optK + " " + lastUnsolK + ",time:");
							System.out.println(((System.currentTimeMillis() - lastStepTime) / 1000d) + "s)");
						}

						if (solve)
						{
							solver = solvertmp;
							this.optK = mid;
							iter = solver.getNumberOfRecursiveSteps();
						} else
						{
							lastUnsolK = mid;
						}
					}
					break;

					// for (int i = lastUnsolK + 1; i < this.optK; i++)
					// {
					// System.out.println("test i:" +i+" "+ this.optK + "(" +
					// lastUnsolK + ")");
					// final SearchTreeSolver solvertmp = new
					// SearchTreeSolver(this.dist, i);
					// if (solvertmp.solve(this.KERNELIZATION,
					// this.INTERLEAVING,
					// this.INTERLEAVE_LP))
					// {
					// solver = solvertmp;
					// this.optK = i;
					// iter = solver.getNumberOfRecursiveSteps();
					// break;
					// }
					// }
					// break;
				}
				lastStepTime = (System.currentTimeMillis() - lastStepTime) / 1000d;
				lastUnsolK = this.optK;
			}
		} else
		{
			this.optK = lowerBound;
			iter = 0;
		}

		this.endTime = (System.currentTimeMillis() - begin) / 1000d;
		this.result = new BenchResult(this.optK, this.endTime, iter, this.result.getLPVal(), this.result.getLpTimeInSec());
	}

	private static void waitUntilFirstResultIsAvailable(Future<Store<Integer, Boolean>> lpResult, ArrayList<Future<Integer>> greedyResults)
	{
		int sleepMilis = 25;
		while (!lpResult.isDone() && !isOneGreedyDone(greedyResults))// wait for the first result
		{
			try
			{
				Thread.sleep(sleepMilis);
			} catch (InterruptedException e)
			{
			}

			sleepMilis = Math.min(100, (int) (sleepMilis * 1.5));
		}

	}

	private static boolean isOneGreedyDone(ArrayList<Future<Integer>> greedyResults)
	{
		for (Future<Integer> result : greedyResults)
		{
			if (result.isDone())
				return true;
		}
		return false;
	}
}