package dataReductionRules;

import solver.hierarchicalClustering.SearchTreeSolver;
import distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction;

/**
 * This rules solves "isolated" conflicts, that means, if there is a conflict such that the corresponding edges only occur in one conflict, this
 * conflict is isolated and can be solved arbitrarily.
 * 
 * @author sepphar
 * 
 */
public class Rule5 extends AdvancedRule
{

	public Rule5(final int interleavingConst)
	{
		super(interleavingConst);
	}

	@Override
	public boolean performReductionRule(final ChangeLogger solver) throws UnsolvableException
	{
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction conflictEdge = solver.getConflictEdgeStatistic();
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction nonConflictEdge = solver.getNonConflictEdgeStatistic();
		boolean repeat = true;
		while (repeat)
		{
			repeat = false;
			for (final int[] conflict : solver.getConflicts())
			{
				final int i = conflict[0];
				final int j = conflict[1];
				final int l = conflict[2];
				final int sum = conflictEdge.get(i, j) + conflictEdge.get(i, l) + conflictEdge.get(j, l) + nonConflictEdge.get(i, j)
						+ nonConflictEdge.get(i, l) + nonConflictEdge.get(j, l);
				if (sum == 3)
				{
					solver.changeDistanceValue(
							i,
							j,
							-1
									* (solver.getDistanceFunction().get(i, j) - Math.max(solver.getDistanceFunction().get(i, l), solver
											.getDistanceFunction().get(j, l))), true);
					repeat = true;
					if (SearchTreeSolver.RulesVerboseLevel > 0)
					{
						System.out.println("rule5_" + solver.getParameterValue());
						System.out.println(solver.getDistanceFunction().get(i, j) + "\t" + solver.getDistanceFunction().get(i, l) + "\t"
								+ solver.getDistanceFunction().get(j, l));
					}
					break;
				}
			}
		}

		return false;
	}
}
