package dataReductionRules.deprecated;

import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import org.jgrapht.alg.BellmanFordShortestPath;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.SimpleGraph;

import dataReductionRules.ChangeLogger;
import dataReductionRules.RuleInterface;
import dataReductionRules.UnsolvableException;
import dataReductionRules.rule3HelpClasses.deprecated.Node;
import dataReductionRules.rule3HelpClasses.deprecated.findCCGraph;
import distanceFunction.SymmetricDiscreteDistanceFunction;

/**
 * Rule: Crtitcal Clique Rule
 * 
 * @author Johannes
 * 
 */

public class Rule3 implements RuleInterface
{

	public static final String ruleKey = "Critical Clique Rule";

	public static SimpleGraph<Node, DefaultEdge> buildTresholdGraph(final SymmetricDiscreteDistanceFunction distanceFunction,
			final Vector<Integer> indices, final int t)
	{
		final boolean verbose = false;

		final SimpleGraph<Node, DefaultEdge> G = new SimpleGraph<Node, DefaultEdge>(DefaultEdge.class);

		// build treshold graph

		// - create vertices
		if (verbose)
		{
			System.err.println("Create verts");
		}
		final Node[] V = new Node[indices.size()];
		for (int j = 0; j < indices.size(); j++)
		{
			V[j] = new Node(indices.get(j), G);
			G.addVertex(V[j]);
		}

		// determine edges
		if (verbose)
		{
			System.out.println("Create edges");
		}
		for (int j = 0; j < V.length; ++j)
		{
			for (int k = j + 1; k < V.length; ++k)
			{
				if (distanceFunction.get(V[j].i, V[k].i) < t)
				{
					G.addEdge(V[j], V[k]);
				}
			}
		}

		if (verbose)
		{
			System.out.println("n.o.v.: " + G.vertexSet().size());
			System.out.println("n.o.e.: " + G.edgeSet().size());
		}
		return G;
	}

	public final boolean debug;

	private int intCount = 0;

	public ChangeLogger solver;

	public Rule3()
	{
		this(false);
	}

	public Rule3(final boolean d)
	{
		this.debug = d;
		this.solver = null;
	}

	public String getKeyword()
	{
		return Rule3.ruleKey;
	}

	@Override
	public boolean performRule(final ChangeLogger solver) throws UnsolvableException
	{

		if ((this.intCount++ % 3) == 0)
		{
			this.solver = solver;
			return this.performRule(solver.getDistanceFunction());
		}
		return false;
	}

	public boolean performRule(final SymmetricDiscreteDistanceFunction distanceFunction) throws UnsolvableException
	{

		// final DistanceFunction distanceFunction =
		// solver.getDistanceFunction();
		final int numberOfElements = distanceFunction.getNoOfElements();

		// Determine "active" indices
		final Vector<Integer> indices = new Vector<Integer>();

		for (int j = 0; j < numberOfElements; j++)
		{
			if (!distanceFunction.isDeletedData(j))
			{
				indices.add(j);
			}
		}

		if (indices.size() == 0)
			return false;

		return this.performRule(distanceFunction, indices);
	}

	// reduce matrix induced by the elements of "indices"
	private boolean performRule(final SymmetricDiscreteDistanceFunction distanceFunction, final Vector<Integer> indices) throws UnsolvableException
	{

		// Test whether indices induce an ultrametric
		// if yes, then delte indices in "indices".
		if (distanceFunction.isUltrametric(indices))
		{
			for (int i = 0; i < indices.size(); ++i)
			{
				if (this.solver == null)
				{
					distanceFunction.deleteData(indices.get(i));
				} else
				{
					this.solver.deleteData(indices.get(i));
				}
			}

			return true;
		}

		// determine Max distance
		int Max = 0;
		for (int j = 0; j < indices.size(); j++)
		{
			for (int t = j + 1; t < indices.size(); t++)
			{
				Max = distanceFunction.get(indices.get(j), indices.get(t)) > Max ? distanceFunction.get(indices.get(j), indices.get(t)) : Max;
			}
		}

		// if(solver!=null)
		// if(indices.size() < Max * 6 * solver.getParameterValue() )
		// {
		// System.out.println("P1");
		// return false;
		// }

		// build treshold graph
		final SimpleGraph<Node, DefaultEdge> G = Rule3.buildTresholdGraph(distanceFunction, indices, Max);

		// Debug, print out adjacency list
		if (this.debug)
		{
			final Iterator<DefaultEdge> E = G.edgeSet().iterator();
			while (E.hasNext())
			{
				final DefaultEdge e = E.next();
				System.out.println(e);
			}
		}

		// Construct critical clique graph
		final SimpleGraph<TreeSet<Node>, DefaultEdge> CC = findCCGraph.ccgraph(G);
		// if ( this.solver != null )
		// {
		// if ( CC.vertexSet().size() > 4 * this.solver.getParameterValue() )
		// {
		// System.out.println("two many cc's");
		// throw new UnsolvableException();
		// }
		// }

		// Find isolated Cliques
		// - correspond to vertices of degree 0
		// ->> recursively reduce instance induced by the indices of the
		// isolated cliques
		boolean reduced = false;
		Iterator<TreeSet<Node>> VCC;
		VCC = CC.vertexSet().iterator();
		while (VCC.hasNext())
		{
			final TreeSet<Node> v = VCC.next();
			if (CC.degreeOf(v) == 0)
			{
				final Vector<Integer> VI = new Vector<Integer>();
				final Iterator<Node> ni = v.iterator();
				while (ni.hasNext())
				{
					final Node vv = ni.next();
					VI.add(vv.i);
				}
				// Recursively reduce instance
				if (this.performRule(distanceFunction, VI))
				{
					reduced = true;
				}
			}
		}

		// Apply Critical Clique Rule
		if (this.ruleCC(CC, distanceFunction, Max))
		{
			reduced = true;
		}

		return reduced;
	}

	public boolean performRuleSubinst(final SymmetricDiscreteDistanceFunction distanceFunction, final Vector<Integer> X) throws UnsolvableException
	{
		// Determine "active" indices
		final Vector<Integer> indices = new Vector<Integer>();

		for (int i = 0; i < X.size(); ++i)
		{
			if (!distanceFunction.isDeletedData(X.get(i)))
			{
				indices.add(X.get(i));
			}
		}

		if (indices.size() == 0)
			return false;

		return this.performRule(distanceFunction, indices);
	}

	public boolean ruleCC(final SimpleGraph<TreeSet<Node>, DefaultEdge> CC, final SymmetricDiscreteDistanceFunction distanceFunction, final int Max)
			throws UnsolvableException
	{

		final Iterator<TreeSet<Node>> VCC = CC.vertexSet().iterator();

		// Check for every vertex, whether the cc-rule applies
		while (VCC.hasNext())
		{
			final TreeSet<Node> s = VCC.next();
			if (CC.degreeOf(s) != 0)
			{
				// Determine first and second neighborhood of s
				final Vector<TreeSet<Node>> N1 = new Vector<TreeSet<Node>>();// first
				// neighborhood
				int n1 = 0;
				final Vector<TreeSet<Node>> N2 = new Vector<TreeSet<Node>>();// second
				// neighborhood
				int n2 = 0;
				final BellmanFordShortestPath<TreeSet<Node>, DefaultEdge> BF = new BellmanFordShortestPath<TreeSet<Node>, DefaultEdge>(CC, s, 2);
				final Iterator<TreeSet<Node>> VCC2 = CC.vertexSet().iterator();
				while (VCC2.hasNext())
				{
					final TreeSet<Node> t = VCC2.next();
					if (t != s)
					{
						if (BF.getCost(t) == 1)
						{
							N1.add(t);
							n1 = n1 + t.size();
						} else if (BF.getCost(t) == 2)
						{
							N2.add(t);
							n2 = n2 + t.size();
						}
					}
				}

				// reduce instance
				// if ( s.size() >= (Max - 1) * (n1 + n2) ) // Instance can be
				// reduced
				if ((s.size() >= ((Max - 1) * n1)) && (s.size() >= (n1 + ((Max - 1) * n2)))) // Instance
				// can
				// be
				// reduced
				{
					int dec = 0;
					// add all missing edges in the the closed neighborhood of s
					final Iterator<Node> is = s.iterator();

					// - between s and N(s)
					while (is.hasNext())
					{
						final Node vs = is.next();
						// newNode.add(vs);
						for (int i = 0; i < N1.size(); ++i)
						{
							final Iterator<Node> st = N1.get(i).iterator();
							// CC.removeVertex(N1.get(i));
							while (st.hasNext())
							{
								final Node vt = st.next();
								// newNode.add(vt);
								if (distanceFunction.get(vs.i, vt.i) == Max)
								{
									if (this.solver == null)
									{
										distanceFunction.set(vs.i, vt.i, Max - 1);
									} else
									{
										this.solver.changeDistanceValue(vs.i, vt.i, -1, true);
									}
									++dec;
								}
							}

						}
					}

					// - between two nodes in N(s)
					for (int i = 0; i < N1.size(); ++i)
					{
						for (int j = i + 1; j < N1.size(); ++j)
						{
							final Iterator<Node> n11 = N1.get(i).iterator();
							while (n11.hasNext())
							{
								final Node v1 = n11.next();
								final Iterator<Node> n12 = N1.get(j).iterator();
								while (n12.hasNext())
								{
									final Node v2 = n12.next();
									if (distanceFunction.get(v1.i, v2.i) == Max)
									{
										if (this.solver == null)
										{
											distanceFunction.set(v1.i, v2.i, Max - 1);
										} else
										{
											this.solver.changeDistanceValue(v1.i, v2.i, -1, true);
										}
										++dec;
									}
								}

							}
						}
					}

					// remove edges between first and second neighborhood
					for (int i = 0; i < N1.size(); ++i)
					{
						final Iterator<Node> st = N1.get(i).iterator();
						while (st.hasNext())
						{
							final Node vt = st.next();
							for (int j = 0; j < N2.size(); ++j)
							{
								final Iterator<Node> stt = N2.get(j).iterator();
								while (stt.hasNext())
								{
									final Node vtt = stt.next();
									final int dist = distanceFunction.get(vt.i, vtt.i);
									if (dist < Max)
									{
										dec = (dec + Max) - distanceFunction.get(vt.i, vtt.i);
										if (this.solver == null)
										{
											distanceFunction.set(vt.i, vtt.i, Max);
										} else
										{
											this.solver.changeDistanceValue(vt.i, vtt.i, Max - dist, true);
										}

									}

								}
							}
						}
					}

					// System.out.println("changes: " + dec);

					if (this.solver != null)
					{
						if (this.solver.getParameterValue() < 0)
							throw new UnsolvableException();
					}
					return true;
				}
			}
		}

		return false;
	}

	// public boolean apply1(SimpleGraph<TreeSet<Node>, DefaultEdge> CC,
	// DistanceFunction
	// distanceFunction, int Max)
	// throws UnsolvableExecption
	// {
	//
	// final Iterator<TreeSet<Node>> VCC = CC.vertexSet().iterator();
	//
	// // Check for every vertex, whether the cc-rule applies
	// while (VCC.hasNext())
	// {
	// final TreeSet<Node> s = VCC.next();
	// if ( CC.degreeOf(s) != 0 )
	// {
	// // Determine first and second neighborhood of s
	// final Vector<TreeSet<Node>> N1 = new Vector<TreeSet<Node>>();// first
	// // neighborhood
	// int n1 = 0;
	// final Vector<TreeSet<Node>> N2 = new Vector<TreeSet<Node>>();// second
	// // neighborhood
	// int n2 = 0;
	// final BellmanFordShortestPath<TreeSet<Node>, DefaultEdge> BF = new
	// BellmanFordShortestPath<TreeSet<Node>, DefaultEdge>(
	// CC, s, 2);
	// final Iterator<TreeSet<Node>> VCC2 = CC.vertexSet().iterator();
	// while (VCC2.hasNext())
	// {
	// final TreeSet<Node> t = VCC2.next();
	// if ( t != s )
	// if ( BF.getCost(t) == 1 )
	// {
	// N1.add(t);
	// n1 = n1 + t.size();
	// }
	// else if ( BF.getCost(t) == 2 )
	// {
	// N2.add(t);
	// n2 = n2 + t.size();
	// }
	// }
	//
	// // reduce instance
	// // if ( s.size() >= (Max - 1) * (n1 + n2) ) // Instance can be reduced
	// if ( s.size() >= n1 ) // Instance
	// // can
	// // be
	// // reduced
	// {
	// int dec = 0;
	// // add all missing edges in the the closed neighborhood of s
	// final Iterator<Node> is = s.iterator();
	//
	// // - between s and N(s)
	// while (is.hasNext())
	// {
	// final Node vs = is.next();
	// // newNode.add(vs);
	// for (int i = 0; i < N1.size(); ++i)
	// {
	// final Iterator<Node> st = N1.get(i).iterator();
	// // CC.removeVertex(N1.get(i));
	// while (st.hasNext())
	// {
	// final Node vt = st.next();
	// // newNode.add(vt);
	// if ( distanceFunction.get(vs.i, vt.i) == Max )
	// {
	// if ( this.solver == null )
	// distanceFunction.set(vs.i, vt.i, Max - 1);
	// else
	// this.solver.changeDistanceValue(vs.i, vt.i, -1, true);
	// ++dec;
	// }
	// }
	//
	// }
	// }
	//
	// // - between two nodes in N(s)
	// for (int i = 0; i < N1.size(); ++i)
	// for (int j = i + 1; j < N1.size(); ++j)
	// {
	// final Iterator<Node> n11 = N1.get(i).iterator();
	// while (n11.hasNext())
	// {
	// final Node v1 = n11.next();
	// final Iterator<Node> n12 = N1.get(j).iterator();
	// while (n12.hasNext())
	// {
	// final Node v2 = n12.next();
	// if ( distanceFunction.get(v1.i, v2.i) == Max )
	// {
	// if ( this.solver == null )
	// distanceFunction.set(v1.i, v2.i, Max - 1);
	// else
	// this.solver.changeDistanceValue(v1.i, v2.i, -1, true);
	// ++dec;
	// }
	// }
	//
	// }
	// }
	//
	// // remove edges between first and second neighborhood
	// for (int i = 0; i < N1.size(); ++i)
	// {
	// final Iterator<Node> st = N1.get(i).iterator();
	// while (st.hasNext())
	// {
	// final Node vt = st.next();
	// for (int j = 0; j < N2.size(); ++j)
	// {
	// final Iterator<Node> stt = N2.get(j).iterator();
	// while (stt.hasNext())
	// {
	// final Node vtt = stt.next();
	// final int dist = distanceFunction.get(vt.i, vtt.i);
	// if ( dist < Max )
	// {
	// dec = dec + Max - distanceFunction.get(vt.i, vtt.i);
	// if ( this.solver == null )
	// distanceFunction.set(vt.i, vtt.i, Max);
	// else
	// this.solver.changeDistanceValue(vt.i, vtt.i, Max - dist, true);
	//
	// }
	//
	// }
	// }
	// }
	// }
	//
	// //System.out.println("changes: " + dec);
	//
	// if ( this.solver != null )
	// {
	// if ( this.solver.getParameterValue() < 0 )
	// throw new UnsolvableExecption();
	// }
	// // if(dec==0)
	// // {
	// // System.err.println(s.size() + " : " + s);
	// // System.err.println(N1.get(0).size() + " : "+N1.get(0));
	// // System.err.println(s.first().key.characters);
	// // System.err.println(N1.get(0).first().key.characters);
	// // System.err.println( N1.get(0).last().key.characters);
	// // System.err.println(G.degreeOf(s.first()));
	// // System.err.println(G.degreeOf(N1.get(0).first()));
	// //
	// // Node v=N1.get(0).first();
	// // System.err.println(G.edgesOf(v));
	// //
	// //
	// //
	// // //VisTools.show(G);
	// // VisTools.show(CC);
	// // while(true)
	// // ;
	// // }
	// return true;
	// }
	//
	// }
	// }
	//
	// return false;
	// }

	// public boolean apply2(SimpleGraph<TreeSet<Node>, DefaultEdge> CC,
	// DistanceFunction
	// distanceFunction, int Max)
	// throws UnsolvableExecption
	// {
	//
	// final Iterator<TreeSet<Node>> VCC = CC.vertexSet().iterator();
	//
	// // Check for every vertex, whether the cc-rule applies
	// while (VCC.hasNext())
	// {
	// final TreeSet<Node> s = VCC.next();
	// if ( CC.degreeOf(s) != 0 )
	// {
	// // Determine first and second neighborhood of s
	// final Vector<TreeSet<Node>> N1 = new Vector<TreeSet<Node>>();// first
	// // neighborhood
	// int n1 = 0;
	// final Vector<TreeSet<Node>> N2 = new Vector<TreeSet<Node>>();// second
	// // neighborhood
	// int n2 = 0;
	// final BellmanFordShortestPath<TreeSet<Node>, DefaultEdge> BF = new
	// BellmanFordShortestPath<TreeSet<Node>, DefaultEdge>(
	// CC, s, 2);
	// final Iterator<TreeSet<Node>> VCC2 = CC.vertexSet().iterator();
	// while (VCC2.hasNext())
	// {
	// final TreeSet<Node> t = VCC2.next();
	// if ( t != s )
	// if ( BF.getCost(t) == 1 )
	// {
	// N1.add(t);
	// n1 = n1 + t.size();
	// }
	// else if ( BF.getCost(t) == 2 )
	// {
	// N2.add(t);
	// n2 = n2 + t.size();
	// }
	// }
	//
	// // reduce instance
	// // if ( s.size() >= (Max - 1) * (n1 + n2) ) // Instance can be reduced
	// if ( 3 * s.size() >= n1 ) // Instance
	// // can
	// // be
	// // reduced
	// {
	// int dec = 0;
	// // add all missing edges in the the closed neighborhood of s
	// final Iterator<Node> is = s.iterator();
	//
	// // - between s and N(s)
	// while (is.hasNext())
	// {
	// final Node vs = is.next();
	// // newNode.add(vs);
	// for (int i = 0; i < N1.size(); ++i)
	// {
	// final Iterator<Node> st = N1.get(i).iterator();
	// // CC.removeVertex(N1.get(i));
	// while (st.hasNext())
	// {
	// final Node vt = st.next();
	// // newNode.add(vt);
	// if ( distanceFunction.get(vs.i, vt.i) == Max )
	// {
	// if ( this.solver == null )
	// distanceFunction.set(vs.i, vt.i, Max - 1);
	// else
	// this.solver.changeDistanceValue(vs.i, vt.i, -1, true);
	// ++dec;
	// }
	// }
	//
	// }
	// }
	//
	// // - between two nodes in N(s)
	// for (int i = 0; i < N1.size(); ++i)
	// for (int j = i + 1; j < N1.size(); ++j)
	// {
	// final Iterator<Node> n11 = N1.get(i).iterator();
	// while (n11.hasNext())
	// {
	// final Node v1 = n11.next();
	// final Iterator<Node> n12 = N1.get(j).iterator();
	// while (n12.hasNext())
	// {
	// final Node v2 = n12.next();
	// if ( distanceFunction.get(v1.i, v2.i) == Max )
	// {
	// if ( this.solver == null )
	// distanceFunction.set(v1.i, v2.i, Max - 1);
	// else
	// this.solver.changeDistanceValue(v1.i, v2.i, -1, true);
	// ++dec;
	// }
	// }
	//
	// }
	// }
	//
	// // remove edges between first and second neighborhood
	// for (int i = 0; i < N1.size(); ++i)
	// {
	// final Iterator<Node> st = N1.get(i).iterator();
	// while (st.hasNext())
	// {
	// final Node vt = st.next();
	// for (int j = 0; j < N2.size(); ++j)
	// {
	// final Iterator<Node> stt = N2.get(j).iterator();
	// while (stt.hasNext())
	// {
	// final Node vtt = stt.next();
	// final int dist = distanceFunction.get(vt.i, vtt.i);
	// if ( dist < Max )
	// {
	// dec = dec + Max - distanceFunction.get(vt.i, vtt.i);
	// if ( this.solver == null )
	// distanceFunction.set(vt.i, vtt.i, Max);
	// else
	// this.solver.changeDistanceValue(vt.i, vtt.i, Max - dist, true);
	//
	// }
	//
	// }
	// }
	// }
	// }
	//
	// System.out.println("changes: " + dec);
	//
	// // if(solver!=null)
	// // {
	// // if(solver.getParameterValue()<0)
	// // throw new UnsolvableExecption();
	// // }
	// // if(dec==0)
	// // {
	// // System.err.println(s.size() + " : " + s);
	// // System.err.println(N1.get(0).size() + " : "+N1.get(0));
	// // System.err.println(s.first().key.characters);
	// // System.err.println(N1.get(0).first().key.characters);
	// // System.err.println( N1.get(0).last().key.characters);
	// // System.err.println(G.degreeOf(s.first()));
	// // System.err.println(G.degreeOf(N1.get(0).first()));
	// //
	// // Node v=N1.get(0).first();
	// // System.err.println(G.edgesOf(v));
	// //
	// //
	// //
	// // //VisTools.show(G);
	// // VisTools.show(CC);
	// // while(true)
	// // ;
	// // }
	// return true;
	// }
	//
	// }
	// }
	//
	// return false;
	// }

	// public boolean apply3(SimpleGraph<TreeSet<Node>, DefaultEdge> CC,
	// DistanceFunction
	// distanceFunction, int Max)
	// throws UnsolvableExecption
	// {
	//
	// final Iterator<TreeSet<Node>> VCC = CC.vertexSet().iterator();
	// final Random random = new Random();
	// final int n = CC.vertexSet().size();
	// final int r = random.nextInt(n);
	//
	// System.out.println(r);
	// // Check for every vertex, whether the cc-rule applies
	// while (VCC.hasNext())
	// {
	// final TreeSet<Node> s = VCC.next();
	// if ( (CC.degreeOf(s) != 0) && random.nextBoolean() )
	// {
	// // Determine first and second neighborhood of s
	// final Vector<TreeSet<Node>> N1 = new Vector<TreeSet<Node>>();// first
	// // neighborhood
	// int n1 = 0;
	// final Vector<TreeSet<Node>> N2 = new Vector<TreeSet<Node>>();// second
	// // neighborhood
	// int n2 = 0;
	// final BellmanFordShortestPath<TreeSet<Node>, DefaultEdge> BF = new
	// BellmanFordShortestPath<TreeSet<Node>, DefaultEdge>(
	// CC, s, 2);
	// final Iterator<TreeSet<Node>> VCC2 = CC.vertexSet().iterator();
	// while (VCC2.hasNext())
	// {
	// final TreeSet<Node> t = VCC2.next();
	// if ( t != s )
	// if ( BF.getCost(t) == 1 )
	// {
	// N1.add(t);
	// n1 = n1 + t.size();
	// }
	// else if ( BF.getCost(t) == 2 )
	// {
	// N2.add(t);
	// n2 = n2 + t.size();
	// }
	// }
	//
	// // reduce instance
	// // if ( s.size() >= (Max - 1) * (n1 + n2) ) // Instance can be reduced
	// // if (s.size() >= n1) // Instance
	// // can
	// // be
	// // reduced
	// {
	// int dec = 0;
	// // add all missing edges in the the closed neighborhood of s
	// final Iterator<Node> is = s.iterator();
	//
	// // - between s and N(s)
	// while (is.hasNext())
	// {
	// final Node vs = is.next();
	// // newNode.add(vs);
	// for (int i = 0; i < N1.size(); ++i)
	// {
	// final Iterator<Node> st = N1.get(i).iterator();
	// // CC.removeVertex(N1.get(i));
	// while (st.hasNext())
	// {
	// final Node vt = st.next();
	// // newNode.add(vt);
	// if ( distanceFunction.get(vs.i, vt.i) == Max )
	// {
	// if ( this.solver == null )
	// distanceFunction.set(vs.i, vt.i, Max - 1);
	// else
	// this.solver.changeDistanceValue(vs.i, vt.i, -1, true);
	// ++dec;
	// }
	// }
	//
	// }
	// }
	//
	// // - between two nodes in N(s)
	// for (int i = 0; i < N1.size(); ++i)
	// for (int j = i + 1; j < N1.size(); ++j)
	// {
	// final Iterator<Node> n11 = N1.get(i).iterator();
	// while (n11.hasNext())
	// {
	// final Node v1 = n11.next();
	// final Iterator<Node> n12 = N1.get(j).iterator();
	// while (n12.hasNext())
	// {
	// final Node v2 = n12.next();
	// if ( distanceFunction.get(v1.i, v2.i) == Max )
	// {
	// if ( this.solver == null )
	// distanceFunction.set(v1.i, v2.i, Max - 1);
	// else
	// this.solver.changeDistanceValue(v1.i, v2.i, -1, true);
	// ++dec;
	// }
	// }
	//
	// }
	// }
	//
	// // remove edges between first and second neighborhood
	// for (int i = 0; i < N1.size(); ++i)
	// {
	// final Iterator<Node> st = N1.get(i).iterator();
	// while (st.hasNext())
	// {
	// final Node vt = st.next();
	// for (int j = 0; j < N2.size(); ++j)
	// {
	// final Iterator<Node> stt = N2.get(j).iterator();
	// while (stt.hasNext())
	// {
	// final Node vtt = stt.next();
	// final int dist = distanceFunction.get(vt.i, vtt.i);
	// if ( dist < Max )
	// {
	// dec = dec + Max - distanceFunction.get(vt.i, vtt.i);
	// if ( this.solver == null )
	// distanceFunction.set(vt.i, vtt.i, Max);
	// else
	// this.solver.changeDistanceValue(vt.i, vtt.i, Max - dist, true);
	//
	// }
	//
	// }
	// }
	// }
	// }
	//
	// System.out.println("changes: " + dec);
	//
	// // if(solver!=null)
	// // {
	// // if(solver.getParameterValue()<0)
	// // throw new UnsolvableExecption();
	// // }
	// // if(dec==0)
	// // {
	// // System.err.println(s.size() + " : " + s);
	// // System.err.println(N1.get(0).size() + " : "+N1.get(0));
	// // System.err.println(s.first().key.characters);
	// // System.err.println(N1.get(0).first().key.characters);
	// // System.err.println( N1.get(0).last().key.characters);
	// // System.err.println(G.degreeOf(s.first()));
	// // System.err.println(G.degreeOf(N1.get(0).first()));
	// //
	// // Node v=N1.get(0).first();
	// // System.err.println(G.edgesOf(v));
	// //
	// //
	// //
	// // //VisTools.show(G);
	// // VisTools.show(CC);
	// // while(true)
	// // ;
	// // }
	// return true;
	// }
	//
	// }
	// }
	//
	// return false;
	// }

	// public boolean heu(DistanceFunction distanceFunction) throws
	// UnsolvableExecption
	// {
	//
	// // final DistanceFunction distanceFunction =
	// solver.getDistanceFunction();
	// final int numberOfElements = distanceFunction.getNoOfElements();
	//
	// // Determine "active" indices
	// final Vector<Integer> indices = new Vector<Integer>();
	//
	// for (int j = 0; j < numberOfElements; j++)
	// if ( !distanceFunction.isDeletedData[j] )
	// indices.add(j);
	//
	// if ( indices.size() == 0 )
	// return false;
	//
	// return this.heu(distanceFunction, indices);
	// }

}
