/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package dataReductionRules.rule3HelpClasses;

/**
 * MicroGraph is a minimal graph implementation, for which vertices are integers.
 * 
 * @author Tomasz Przedmojski
 */
public class MicroGraph
{
	protected SparseIntegerSet[] adjacencySet;
	protected int maxElement;

	/**
	 * Constructs instance of MicroGraph with maximum number of elements in graph passed as parameters.
	 * 
	 * <b>Do not use integers larger than or equal to maxElement with this class!</b> This class does not check passed values, so if you make a
	 * mistake - you die.
	 * 
	 * @param maxElement
	 */
	public MicroGraph(final int maxElement)
	{
		this.maxElement = maxElement;

		this.adjacencySet = new SparseIntegerSet[maxElement];
		for (int i = 0; i < this.maxElement; ++i)
		{
			this.adjacencySet[i] = new SparseIntegerSet(this.maxElement);
		}

	}

	/**
	 * Adds undirected edge between nodes u and v. Nothing happens if the edge is already there.
	 * 
	 * @param u
	 * @param v
	 */
	public void addEdge(final int u, final int v)
	{
		this.adjacencySet[u].add(v);
		this.adjacencySet[v].add(u);
	}

	/**
	 * Returns true if nodes u and v are joined by an edge.
	 * 
	 * @param u
	 * @param v
	 * @return
	 */
	public boolean containsEdge(final int u, final int v)
	{
		return this.adjacencySet[u].contains(v);
	}

	/**
	 * Returns maximum number of nodes in this graph.
	 * 
	 * @return
	 */
	public int getMaxNodes()
	{
		return this.maxElement;
	}

	/**
	 * Returns a set of neighbors of u (i.e. all nodes v such that there exists and edge between u and v).
	 * 
	 * @param u
	 * @return
	 */
	public SparseIntegerSet neighborsOf(final int u)
	{
		return this.adjacencySet[u];
	}

	/**
	 * Removes undirected edge between nodes u and v. Nothing happens if the edge hasn't been there.
	 * 
	 * @param u
	 * @param v
	 */
	public void removeEdge(final int u, final int v)
	{
		this.adjacencySet[u].remove(v);
		this.adjacencySet[v].remove(u);
	}
}
