package solver;

import java.util.Collection;
import java.util.Collections;
import java.util.Vector;

import dataReductionRules.ObservingAdvancedRule;
import dataReductionRules.RuleInterface;
import dataReductionRules.UnsolvableException;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public abstract class Solver extends SymmetricDiscreteDistanceFunction
{
	private int parameter;

	private final Vector<RuleInterface> rules;

	public Solver(final SymmetricDiscreteDistanceFunction distanceFunction, final int k)
	{
		super(distanceFunction);
		this.parameter = k;
		this.rules = new Vector<RuleInterface>();

	}

	public void addObservingRule(final ObservingAdvancedRule rule)
	{
		this.addRule(rule);
		this.addObserver(rule);
	}

	public void addRule(final RuleInterface rule)
	{
		this.rules.add(rule);
	}

	protected void decreaseParameter()
	{
		this.decreaseParameter(1);
	}

	protected void decreaseParameter(final int distance)
	{
		this.parameter -= distance;

	}

	public int getParameterValue()
	{
		return this.parameter;
	}

	public Collection<RuleInterface> getRules()
	{
		return Collections.unmodifiableCollection(this.rules);
	}

	protected void increaseParameter()
	{
		this.increaseParameter(1);
	}

	protected void increaseParameter(final int distance)
	{
		this.parameter += distance;
	}

	public void randomizeRuleOrder()
	{
		Collections.shuffle(this.rules);
	}

	public abstract boolean solve() throws UnsolvableException;
}
