package solver.heuristics;

import org.apache.commons.lang3.mutable.MutableDouble;
import org.apache.commons.lang3.mutable.MutableFloat;
import org.apache.commons.lang3.mutable.MutableInt;

import solver.hierarchicalClustering.Conflicts;
import solver.hierarchicalClustering.HierarchicalClusteringSolver;
import sun.reflect.generics.reflectiveObjects.NotImplementedException;
import ProgOpt.SmacClusterEditOpt;
import dataReductionRules.Rule6;
import dataReductionRules.UnsolvableException;
import distanceFunction.ReadOnlyViewOnSymmetricDiscreteDistanceFunction;
import distanceFunction.SymmetricDiscreteDistanceFunction;

public class GreedyUpperBoundHeuristic extends HierarchicalClusteringSolver
{
	public static final MutableDouble RANDOM_ADD_COMPARISON = new MutableDouble(2f);

	public static final MutableInt Max_Queue_Size = new MutableInt(1000);

	public static final MutableDouble RANDOM_SCORE_ADD = new MutableDouble(3d);

	class ConflictScore
	{

		public final int i;
		public final int j;
		public final int l;
		public final int possibleChanges;
		private final int score;
		public final int score_i_j;
		public final int score_i_l;
		public final int score_j_l;

		public ConflictScore(final int[] con, final ReadOnlyViewOnSymmetricDiscreteDistanceFunction conflictEdge,
				final ReadOnlyViewOnSymmetricDiscreteDistanceFunction nonConflictEdge)
		{
			this.i = con[0];
			this.j = con[1];
			this.l = con[2];
			this.possibleChanges = (GreedyUpperBoundHeuristic.this.modStatistic.isDecreasePossible(this.i, this.j) ? 1 : 0)
					+ (GreedyUpperBoundHeuristic.this.modStatistic.isIncreasePossible(this.i, this.l) ? 1 : 0)
					+ (GreedyUpperBoundHeuristic.this.modStatistic.isIncreasePossible(this.j, this.l) ? 1 : 0);
			this.score_i_j = GreedyUpperBoundHeuristic.this.modStatistic.isDecreasePossible(this.i, this.j) ? Math.max(
					conflictEdge.get(this.i, this.j), nonConflictEdge.get(this.i, this.j)) : 0;
			this.score_i_l = GreedyUpperBoundHeuristic.this.modStatistic.isIncreasePossible(this.i, this.l) ? Math.max(
					conflictEdge.get(this.i, this.l), nonConflictEdge.get(this.i, this.l)) : 0;
			this.score_j_l = GreedyUpperBoundHeuristic.this.modStatistic.isIncreasePossible(this.j, this.l) ? Math.max(
					conflictEdge.get(this.j, this.l), nonConflictEdge.get(this.j, this.l)) : 0;
			// this.score = Math.max(Math.max(score_i_j, score_i_l), score_j_l)
			// + (int) Math.round(Math.random() * 2f);
			this.score = Math.max(Math.max(this.score_i_j, this.score_i_l), this.score_j_l)
					+ (int) Math.round(Math.random() * RANDOM_SCORE_ADD.floatValue());

			// this.score = score_i_j + score_i_l + score_j_l + (int)
			// Math.round(Math.random() * 3f);
		}

		public int getMax_Ms()
		{
			return this.score;
		}

		public boolean isSomethingFixed()
		{
			return this.possibleChanges < 3;
		}
	}

	public static final MutableInt GreedyheuristicUpperBoundRuns = new MutableInt(10);

	public final static MutableDouble upperBoundTimeProportion = new MutableDouble(0.08d);

	public static int verboseLevel = 0;

	public static int computeUpperBound(final SymmetricDiscreteDistanceFunction dist)
	{
		return GreedyUpperBoundHeuristic.computeUpperBound(dist, null);
	}

	public static int computeUpperBound(final SymmetricDiscreteDistanceFunction dist, final Conflicts conflicts)
	{
		int upperBound = Integer.MAX_VALUE;
		final double maxTime = GreedyUpperBoundHeuristic.upperBoundTimeProportion.doubleValue() * SmacClusterEditOpt.cutoff_time.doubleValue();
		final long t = System.currentTimeMillis();

		for (int i = 0; i < GreedyUpperBoundHeuristic.GreedyheuristicUpperBoundRuns.intValue(); i++)
		{
			upperBound = Math.min(new GreedyUpperBoundHeuristic(dist, conflicts == null ? null : conflicts.clone()).getUpperBound(), upperBound);

			if (((System.currentTimeMillis() - t) / 1000d) > maxTime)
			{
				break;
			}
		}

		return upperBound;
	}

	private final SymmetricDiscreteDistanceFunction orignalDist;

	private GreedyUpperBoundHeuristic(final SymmetricDiscreteDistanceFunction dist)
	{
		this(dist, null);
	}

	private GreedyUpperBoundHeuristic(final SymmetricDiscreteDistanceFunction dist, final Conflicts conflicts)
	{
		super(dist, -1, conflicts);
		this.orignalDist = new SymmetricDiscreteDistanceFunction(dist);
		this.addRule(new Rule6(1));
	}

	private void computeUpperBound() throws UnsolvableException
	{
		// int mod = 0;
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction conflictEdge = this.getConflictStorage().getConflictEdgeStatistic();
		final ReadOnlyViewOnSymmetricDiscreteDistanceFunction nonConflictEdge = this.getConflictStorage().getNonConflictEdgeStatistic();
		this.performDataReductionRules(false);

		// {// first solve all independent conflicts
		// final Vector<int[]> independentConflicts = Rule0.getIndependentConflicts(this);
		// for (final int[] con : independentConflicts)
		// {
		// this.performBestChange(new ConflictScore(con, conflictEdge, nonConflictEdge));
		// }
		// this.performDataReductionRules(false);
		// }

		while (!this.getConflictStorage().isEmpty())
		{
			ConflictScore bestCon = null;
			{
				int count = 1;
				for (final int[] con : this.getConflictStorage().getConflicts())
				{
					if (count++ > GreedyUpperBoundHeuristic.Max_Queue_Size.intValue())
						break;
					final ConflictScore conStore = new ConflictScore(con, conflictEdge, nonConflictEdge);

					if ((bestCon == null) || (!bestCon.isSomethingFixed() && conStore.isSomethingFixed()))
					{
						bestCon = conStore;
					} else if ((!bestCon.isSomethingFixed() || conStore.isSomethingFixed())
							&& ((conStore.getMax_Ms() + (int) Math
									.round(GreedyUpperBoundHeuristic.RANDOM_ADD_COMPARISON.floatValue() * Math.random())) > bestCon.getMax_Ms()))
					{
						bestCon = conStore;
					}
				}
			}

			if (bestCon.possibleChanges == 0)
			{
				break;
			}
			this.performBestChange(bestCon);
			this.performDataReductionRules(false);
		}
	}

	private int getUpperBound()
	{
		try
		{
			this.computeUpperBound();
		} catch (final UnsolvableException e)
		{
			return Integer.MAX_VALUE;
		}
		if (this.isUltrametric())
			return this.orignalDist.getDistance(this);
		return Integer.MAX_VALUE;
	}

	private void performBestChange(final ConflictScore bestConStore)
	{
		this.performBestChange(bestConStore.i, bestConStore.j, bestConStore.l, bestConStore.score_i_j, bestConStore.score_i_l, bestConStore.score_j_l);
	}

	private void performBestChange(final int i, final int j, final int l, final int m_i_j, final int m_i_l, final int m_j_l)
	{
		if (m_i_j >= Math.max(m_i_l, m_j_l))
		{
			this.changeDistanceValue(i, j, -1, true);
			if (GreedyUpperBoundHeuristic.verboseLevel > 1)
			{
				System.out.println("change:\t" + i + "\t" + j + "\t-1");
			}
		} else if (m_i_l >= Math.max(m_i_j, m_j_l))
		{
			this.changeDistanceValue(i, l, 1, true);
			if (GreedyUpperBoundHeuristic.verboseLevel > 1)
			{
				System.out.println("change:\t" + i + "\t" + l + "\t 1");
			}
		} else
		{
			this.changeDistanceValue(j, l, 1, true);
			if (GreedyUpperBoundHeuristic.verboseLevel > 1)
			{
				System.out.println("change:\t" + j + "\t" + l + "\t 1");
			}
		}
	}

	@Override
	public boolean solve() throws UnsolvableException
	{
		throw new NotImplementedException();
	}

}
